/**
 * JASMINe
 * Copyright (C) 2010 Bull S.A.S.
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * $Id: PurgeDataBase.java 6984 2010-10-04 09:08:56Z jlegrand $
 */
package org.ow2.jasmine.monitoring.commands;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.text.Format;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.commons.cli.AlreadySelectedException;
import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.MissingOptionException;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionGroup;
import org.apache.commons.cli.Options;

import org.ow2.easybeans.component.smartclient.spi.SmartContextFactory;

import org.ow2.jasmine.monitoring.eventswitch.beans.JasmineEventTimerSLBRemote;
import org.ow2.jasmine.event.messages.JasmineEventTimer;


import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
/**
 * Implements 'purge' command, a command that allows JASMINe DB purge.
 * It is based on a StalessSessionBean deployed on the JASMINe's host server.
 * @author danesa
 *
 */
public class PurgeDataBase {

    /**
     * JNDI name of the StalessSessionBean that implements Purge function.
     */
    private final static String EJB_JNDI_NAME = "db-ejb/timer";

    /**
     * Reference to the StalessSessionBean that implements Purge function.
     */
    private static JasmineEventTimerSLBRemote purgeEjb = null;

    /**
     *
     */
    private final static String DATE_FORMAT = "yyyy/MM/dd-HH:mm";

    /**
     *
     */
    private Format formatter;

    /**
     * Logger
     */
    private static Log logger = LogFactory.getLog(PurgeDataBase.class);

    /**
     * Command arguments.
     */
    protected String[] arguments = null;

    /**
     * Name of the command.
     */
    protected String name = "purge";

    /**
     * List of options that should be parsed from the command line.
     */
    private Options options = null;

    /**
     * The command line
     */
    private CommandLine commandLine = null;

    /**
     *
     */
    private long maxEntryAge;

    /**
     * Default value when maxEntryAge is used as criteria.
     */
    private int maxEntryNumber= -1;

    /**
     *
     */
    private Date executionStartDate = null;

    /**
     *
     */
    private long executionInterval;

    /**
     * By default, execute immediately
     * set to false if startDate provided
     */
    private boolean immediateExecution = true;

    /**
     *
     */
    private boolean executeOnAllServers = true;

    /**
     *
     */
    private String domainName;

    /**
     *
     */
    private String serverName;

    /**
     * The constructor allows to initialize the purge command's options
     */
    public PurgeDataBase() {
        this.setOptions();
        formatter = new SimpleDateFormat(DATE_FORMAT);
    }

    /**
     * Define all the possible options for this command.
     */
    private void setOptions() {
        // contains the description of all the possible options
        options = new Options();

        // create an Option group for maxAge and maxNumber
        OptionGroup ageOrNumber = new OptionGroup();
        ageOrNumber.setRequired(true);
        // create the options: maxAge
        Option option = getOption("maxAge");
        ageOrNumber.addOption(option);
        // maxEntries
        option = getOption("maxNumber");
        ageOrNumber.addOption(option);
        options.addOptionGroup(ageOrNumber);

        // server name
        option = getOption("serverName");
        options.addOption(option);
        // domain name
        option = getOption("domainName");
        options.addOption(option);

        // start date
        option = getOption("startDate");
        options.addOption(option);

        // execution period
        option = getOption("execPeriod");
        options.addOption(option);

    }

    /**
     * Allows to create a given option.
     * @param optionName the name of the option
     * @return The created Option object
     */
    private Option getOption(String optionName) {
        Option option = null;
        if (optionName.equals("maxAge")) {
            option = new Option("maxAge", "maxAge", true, "Entries older then maxAge will be purged");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("max entry age");
        } else if (optionName.equals("maxNumber")) {
            option = new Option("maxNumber", "maxNumber", true, "Maximum number of entries that are not purged");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("max entry number");
        } else if (optionName.equals("serverName")) {
            option = new Option("serverName", "serverName", true, "The entries that will be purged are related to this server");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("server name");
        } else if (optionName.equals("domainName")) {
            option = new Option("domainName", "domainName", true, "The entries that will be purged are related to this domain");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("domain name");
        } else if (optionName.equals("startDate")) {
            option = new Option("startDate", "startDate", true, "Date for purge to start");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("start date");
        } else if (optionName.equals("execPeriod")) {
            option = new Option("execPeriod", "execPeriod", true, "Time interval for purge execution");
            option.setRequired(false);
            option.setArgs(1);
            option.setArgName("execution period");
        }
        return option;
    }

    /**
     * Get the command arguments from the command line.
     * @param args
     * @throws ParseException
     * @throws AlreadySelectedException
     * @throws java.text.ParseException
     */
    private void setArgs(final String[] args) throws ParseException, AlreadySelectedException, MissingOptionException, java.text.ParseException {

        BasicParser bp = new BasicParser();
        commandLine = bp.parse(options, args);
        if (commandLine.hasOption("maxAge")) {
            setMaxEntryAge(commandLine.getOptionValue("maxAge"));
        }
        if (commandLine.hasOption("maxNumber")) {
            setMaxEntryNumber(commandLine.getOptionValue("maxNumber"));
        }
        if (commandLine.hasOption("startDate")) {
            setExecutionStartDate(commandLine.getOptionValue("startDate"));
            immediateExecution = false;
        }
        if (commandLine.hasOption("execPeriod")) {
            setExecutionInterval(commandLine.getOptionValue("execPeriod"));
            immediateExecution = false;
        }
        if (commandLine.hasOption("serverName")) {
            setServerName(commandLine.getOptionValue("serverName"));
            executeOnAllServers = false;
        }
        if (commandLine.hasOption("domainName")) {
            setDomainName(commandLine.getOptionValue("domainName"));
            executeOnAllServers = false;
        }

    }

    /**
     * Command execution: call the sole business method defined by the purgeEjb, configureTimer()
     */
    private void exec() {
        JasmineEventTimer purgeTimer = initPurgeTimer();
        try {
            purgeEjb.configureTimer(purgeTimer);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Construct the purgeEjb's configureTimer() method's argument.
     * @return
     */
    private JasmineEventTimer initPurgeTimer() {

        JasmineEventTimer event = new JasmineEventTimer();

        event.setMaxEntryAge(getMaxEntryAge());
        event.setMaxEntryNumber(getMaxEntryNumber());
        event.setExecutionStartDate(getExecutionStartDate());
        //event.setExecutionStartHours(getExecutionStartHours());
        event.setExecutionInterval(getExecutionInterval());
        if (getExecutionInterval() != 0 && getExecutionStartDate() == null) {
            setExecutionStartDate();
            event.setExecutionStartDate(getExecutionStartDate());
        }
        event.setExecuteOnAllServers(isExecuteOnAllServers());
        event.setImmediateExecution(isImmediateExecution());
        event.setDomainName(getDomainName());
        event.setServerName(getServerName());

        return event;
    }

    /**
     * @return maxEntryAge the value of maxEntryAge as set by the -maxAge option
     */
    private long getMaxEntryAge() {
        return maxEntryAge;
    }

    /**
     * Determine maxEntryAge in milliseconds from a String having DURATION_FORMAT.
     * @param maxEntryAge_param -maxAge option
     */
    private void setMaxEntryAge(String maxEntryAge_param) {
        try {
            maxEntryAge = getMilliSeconds(maxEntryAge_param);
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
        }
    }

    /**
     *
     * @return maxEntryNumber value as set by -maxNumber option
     */
    private int getMaxEntryNumber() {
        return maxEntryNumber;
    }

    /**
     * Determine maxEntryNumber
     * @param maxEntryNumber_param the -maxNumber option
     */
    private void setMaxEntryNumber(String maxEntryNumber_param) {
        maxEntryNumber = (new Integer(maxEntryNumber_param)).intValue();
    }

    /**
     * @return the value of executionStartDate as set by the -startDate option
     */
    private Date getExecutionStartDate() {
        return executionStartDate;
    }

    /**
     * Determine executionStartDate from a String having the DATE_FORMAT
     * @param executionStartDate_param the -startDate option
     * @throws java.text.ParseException
     */
    private void setExecutionStartDate(String executionStartDate_param) throws java.text.ParseException {
        Date date = null;
        try {
            date = (Date) formatter.parseObject(executionStartDate_param);
        } catch (java.text.ParseException e) {
            throw e;
        }
        executionStartDate = date;
        System.out.println("Periodic execution starts : " + executionStartDate);
    }

    /**
     * Default start date for periodic execution.
     * This means this is a periodic execution.
     */
    private void setExecutionStartDate() {
        executionStartDate = Calendar.getInstance().getTime();
        System.out.println("Periodic execution starts now : " + executionStartDate);
    }

    private long getExecutionInterval() {
        return executionInterval;
    }


    private void setExecutionInterval(String executionInterval_param) throws IllegalArgumentException {
        executionInterval = getMilliSeconds(executionInterval_param);
    }

    private boolean isExecuteOnAllServers() {
        return executeOnAllServers;
    }

    private boolean isImmediateExecution() {
        return immediateExecution;
    }

    private String getServerName() {
        return serverName;
    }

    private void setServerName(String serverName_param) {
        serverName =  serverName_param;
    }

    private String getDomainName() {
        return domainName;
    }

    private void setDomainName(String domainName_param) {
        domainName = domainName_param;
    }

    /**
     *
     * @param duration String having DURATION_FORMAT.
     * @return milliseconds corresponding to the number of days, hours and minutes.
     * @throws IllegalArgumentException The format is not respected.
     * For example, the days, hours and minutes have to be provided in this order.
     */
    private long getMilliSeconds(String duration) throws IllegalArgumentException {
        duration = duration.toLowerCase();
        String daysMarquer = "d";
        String hoursMarquer = "h";
        String minutesMarquer = "m";
        long numberOfMilliSeconds = 0;
        int numberOfDays = 0;
        int numberOfHours = 0;
        int numberOfMinutes = 0;
        int dayIndex = duration.indexOf(daysMarquer);
        int hoursIndex = duration.indexOf(hoursMarquer);
        int minutesIndex = duration.indexOf(minutesMarquer);
        if (dayIndex == -1 && hoursIndex == -1 && minutesIndex == -1) {
            throw new IllegalArgumentException("Duration " + duration + " is missformed. The correct form is daysDhoursHminutesM");
        }
        // check that order is respected : days - hours - minutes
        boolean orderOk = true;
        if (minutesIndex > 0) {
            if (minutesIndex < dayIndex || minutesIndex < hoursIndex) {
                orderOk = false;
            }
        }
        if (hoursIndex > 0) {
            if (hoursIndex < dayIndex) {
                orderOk = false;
            }
        }
        if (!orderOk) {
            throw new IllegalArgumentException("Duration " + duration + " is missformed. The correct form is daysDhoursHminutesM");
        }
        // look for number of days
        int index = duration.indexOf(daysMarquer);
        if (index > 0) {
            int beginIndex = 0;
            String daysValue = duration.substring(beginIndex, index);
            duration = duration.substring(index + 1, duration.length());
            numberOfDays = new Integer(daysValue) ;
        }
        numberOfHours = 24 * numberOfDays;

        // look for number of hours
        index = duration.indexOf(hoursMarquer);
        if (index > 0) {
            int beginIndex = 0;
            String hoursValue = duration.substring(beginIndex, index);
            duration = duration.substring(index + 1, duration.length());
            numberOfHours = numberOfHours + new Integer(hoursValue);
        }
        numberOfMinutes = 60 * numberOfHours;

        // look for number of minutes
        index = duration.indexOf(minutesMarquer);
        if (index > 0) {
            int beginIndex = 0;
            String minutesValue = duration.substring(beginIndex, index);
            numberOfMinutes = numberOfMinutes + new Integer(minutesValue);
        }
        numberOfMilliSeconds = 60000 * numberOfMinutes;
        return numberOfMilliSeconds;

    }

    /**
     *
     * @throws NamingException
     */
    private static void initEJBConnection() throws NamingException {
        Hashtable<String, Object> env = new Hashtable<String, Object>();
        env.put(Context.INITIAL_CONTEXT_FACTORY, SmartContextFactory.class.getName());
        Context context = new InitialContext(env);
        purgeEjb = (JasmineEventTimerSLBRemote) context.lookup(EJB_JNDI_NAME);
    }


      /**
     * The main
     *
     * @param args command line arguments
     */
    public static void main(final String[] args) {
        PurgeDataBase purge = null;
        try {
            purge = new PurgeDataBase();

            if (args.length <= 1) {
                purge.help();
                System.exit(-1);
            }

            purge.setArgs(args);

            initEJBConnection();

            purge.exec();

        } catch (AlreadySelectedException ase) {
            System.err.println("Error in purge command arguments: " + ase.getMessage());
            purge.help();
        } catch (java.text.ParseException pe) {
            System.err.println("Error in purge command arguments: " + pe.getMessage());
            purge.help();
        } catch (MissingOptionException mi) {
            System.err.println("Error in purge command arguments: " + mi.getMessage());
            purge.help();
        } catch (NamingException ne) {
            System.err.println("Cannot connect to the JASMINe server: " + ne.getMessage());
            ne.printStackTrace();
        } catch (IllegalArgumentException ia) {
            System.err.println("Error in purge command arguments: " + ia.getMessage());
            purge.help();

        } catch (Exception e) {
            e.printStackTrace();
            System.exit(-1);
        }
    }

    public final String getName() {
        return this.name;
    }

    /**
     * Implementation of inherited method.
     *
     * Displays the help for this command.
     *
     * @see org.ow2.jasmine.monitoring.mbeancmd.Command#help()
     * @see AbstractCommand#getHelpDoc()
     */
    public void help() {
        ClassLoader classL = Thread.currentThread().getContextClassLoader();
        InputStream in0 = Thread.currentThread().getContextClassLoader().getResourceAsStream(this.getHelpDoc());

        if (in0 != null) {
            BufferedReader in = new BufferedReader(new InputStreamReader(in0));
            OutputStreamWriter out = new OutputStreamWriter(System.out);
            try {

                boolean goOn = true;
                char[] buf = new char[2048];
                while (goOn) {
                    int read;

                    read = in.read(buf);
                    if (read > 0) {
                        out.write(buf, 0, read);
                        out.flush();
                    } else {
                        goOn = false;
                        System.out.println();
                    }
                }

            } catch (IOException e) {
                logger.error("Error reading help", e);
            } finally {

                if (in != null) {
                    try {
                        in.close();
                    } catch (IOException e) {
                        logger.error("Error closing the BufferedReader", e);
                    }
                }
            }
        } else {
            System.out.println("No help available.");
        }

    }

    /**
     * Gets the help document corresponding to this command.
     *
     * @return Name of the help document corresponding to this command.
     */
    protected String getHelpDoc() {
        String classRadix = this.getClass().getName().substring(this.getClass().getName().lastIndexOf(".") + 1);
        String packName = this.getClass().getPackage().getName();
        String helpdoc = packName.substring(0, packName.lastIndexOf("."));
        helpdoc = helpdoc.replaceAll("[.]", "/") + "/helpdoc/" + classRadix + ".help";
        return helpdoc;
    }


    public String summary() {
        return "Purge JASMINe DataBase";
    }
}
